import random
import typing
import urllib.parse
from datetime import datetime
from timeit import default_timer as timer

from sqlalchemy import Column, Table, MetaData, create_engine

from bakplane.bakplane_pb2 import (
    ResolveResourcePathResponse,
    ResolveResourceIntentResponse,
    ComparisonType,
)
from bakplane.extensions.base import (
    WriteResponse,
    WriteRequest,
    ReadResponse,
    ReadAllRequest,
    ReadRequest,
    ReadRequestBuildingBlocks,
    DataStoreHelper,
)
from bakplane.extensions.spark.utils import (
    to_spark_save_mode,
    timestamp_to_date_string,
    datetime_to_date_string,
    timestamp_to_datetime,
)
from bakplane.utils import to_execution_statistics


class JDBCStoreHelper(DataStoreHelper):
    def __init__(
        self,
        user: str,
        password: str,
        host: str,
        port: int,
        database: str,
        schema_name: str,
        driver: str,
    ):
        self.metadata = MetaData()
        self.engine = create_engine(
            f"{driver}://{user}:{password}@{host}:{port}/{database}"
        )
        self.schema_name = schema_name

    def create_temporary_table(
        self,
        prefix: str,
        columns: typing.List[Column],
        rows: typing.List[typing.Dict[str, typing.Any]],
    ) -> Table:
        tbl = Table(
            prefix
            + "_"
            + str(int(datetime.utcnow().timestamp()))
            + "_"
            + str(random.randint(1, 1000)),
            self.metadata,
            *columns,
            schema=self.schema_name,
        )

        self.metadata.create_all(self.engine)
        self.engine.execute(tbl.insert(), rows)

        return tbl


def __build_read_all_query(
    r: ReadAllRequest, res: ResolveResourcePathResponse
) -> str:
    return f'select {", ".join(r.columns)} from {res.context["schema_name"]}.{res.context["table_name"]}'


def __build_read_query(
    r: ReadRequest,
    res: ResolveResourceIntentResponse,
    dialect,
    store_helper: DataStoreHelper,
) -> str:
    from sqlalchemy import (
        Table,
        MetaData,
        Column,
        select,
        and_,
    )

    metadata: MetaData = MetaData()

    src_tbl: Table = Table(
        res.table_name,
        metadata,
        *[Column(x.name) for x in res.columns],
        schema=res.schema_name,
    )

    mapping_tbl: Table = Table(
        res.mapping.table_name,
        metadata,
        *[Column(x.name) for x in res.mapping.columns],
        schema=res.mapping.schema_name,
    )

    a = src_tbl.alias("a")

    mapping_columns = []
    for column in res.mapping.columns:
        if not column.is_effective_dating:
            mapping_columns.append(
                mapping_tbl.c.get(column.name).label(column.name)
            )

    m = mapping_tbl.alias("m")
    clauses = []
    for clause in res.mapping.clauses:
        if clause.comparison == ComparisonType.EQ:
            clauses.append(
                m.c.get(clause.mapping_field_name)
                == a.c.get(clause.resource_field_name)
            )

    blocks = ReadRequestBuildingBlocks(
        mapping_table=m,
        asset_table=a,
        selectable=m.join(a, and_(*clauses)),
        effective_start_dt=timestamp_to_datetime(
            res.effective_dating.start_dt.seconds
        ),
        effective_end_dt=timestamp_to_datetime(
            res.effective_dating.end_dt.seconds
        ),
        knowledge_start_dt=r.knowledge_start_dt,
        knowledge_end_dt=r.knowledge_end_dt,
        constraints=[],
    )

    if (r.columns is None or len(r.columns) <= 0) and r.query_builder is None:
        raise RuntimeError("You must provide a set of columns.")

    if r.query_builder:
        res = r.query_builder(blocks)
    else:
        blocks.columns = [a.c.get(x) for x in r.columns if x in a.c]
        blocks.constraints = []

        if r.additional_constraints is not None:
            blocks.constraints.append(r.additional_constraints(blocks))

        if r.effective_dating_hint is None:
            blocks.constraints.append(
                a.c.effective_start_dt <= m.c.effective_end_dt
            )
            blocks.constraints.append(
                a.c.effective_end_dt >= m.c.effective_start_dt
            )
            blocks.constraints.append(
                a.c.effective_start_dt
                <= timestamp_to_date_string(
                    res.effective_dating.end_dt.seconds
                )
            )
            blocks.constraints.append(
                a.c.effective_end_dt
                >= timestamp_to_date_string(
                    res.effective_dating.start_dt.seconds
                )
            )
        else:
            if isinstance(r.effective_dating_hint, tuple):
                s = datetime_to_date_string(
                    typing.cast(datetime, r.effective_dating_hint[0])
                )
                e = datetime_to_date_string(
                    typing.cast(datetime, r.effective_dating_hint[1])
                )

                blocks.constraints.append(a.c.effective_start_dt <= e)
                blocks.constraints.append(a.c.effective_end_dt >= s)
            elif isinstance(r.effective_dating_hint, str):
                col = typing.cast(str, r.effective_dating_hint)

                s = col + "_start_dt"
                e = col + "_end_dt"

                blocks.constraints.append(a.c.effective_start_dt <= m.c.get(e))
                blocks.constraints.append(a.c.effective_end_dt >= m.c.get(s))
            else:
                raise RuntimeError("Invalid type for effective_dating_hint.")

        if r.modifiers:
            for m in r.modifiers:
                blocks = m.execute(blocks, store_helper)

        res = (
            select(blocks.columns)
            .select_from(blocks.selectable)
            .where(and_(*blocks.constraints))
        )

    query = str(
        res.compile(
            dialect=dialect,
            compile_kwargs={"literal_binds": True},
        )
    )

    return query


def write_fn(
    r: WriteRequest, res: ResolveResourcePathResponse, driver: str, url: str
) -> WriteResponse:
    start = timer()

    r.df.write.format("jdbc").mode(to_spark_save_mode(r.mode)).option(
        "driver", driver
    ).option("url", url).option(
        "dbtable", f"{res.context['schema_name']}.{res.context['table_name']}"
    ).option(
        "user", res.context["username"]
    ).option(
        "password", urllib.parse.unquote(res.context["password"])
    ).save()

    end = timer()

    return WriteResponse(
        execution_statistics=to_execution_statistics(start, end)
    )


def __read_query(
    spark, url: str, driver: str, query: str, user: str, password: str
):
    start = timer()

    df = (
        spark.read.format("jdbc")
        .option("url", url)
        .option("query", query)
        .option("user", user)
        .option("password", urllib.parse.unquote(password))
        .option("driver", driver)
        .load()
    )

    end = timer()

    return ReadResponse(
        df=df,
        execution_statistics=to_execution_statistics(start, end),
        context={
            "query": query,
        },
    )


def read_all_fn(
    ctx: typing.Any,
    r: ReadAllRequest,
    res: ResolveResourcePathResponse,
    url: str,
    driver: str,
) -> ReadResponse:
    return __read_query(
        spark=ctx,
        url=url,
        driver=driver,
        query=__build_read_all_query(r, res),
        user=res.context["username"],
        password=res.context["password"],
    )


def read_fn(
    ctx: typing.Any,
    r: ReadRequest,
    res: ResolveResourceIntentResponse,
    dialect,
    url: str,
    driver: str,
    store_helper: DataStoreHelper,
) -> ReadResponse:
    return __read_query(
        spark=ctx,
        url=url,
        driver=driver,
        query=__build_read_query(r, res, dialect, store_helper),
        user=res.context["username"],
        password=res.context["password"],
    )



from bakplane.bakplane_pb2 import ErrorEntry
import os
import tempfile
import typing
from abc import abstractmethod, ABC
from typing import IO

from google.protobuf.internal.encoder import _VarintBytes

from bakplane.ingestion.landing_zone import LandingZoneFactory
from .models import Asset, Entity, Relationship, IngestionSessionContext, Error


class Emitter(ABC):
    @abstractmethod
    def emit_assets(self, assets: typing.List[Asset]):
        pass

    @abstractmethod
    def emit_entities(self, entities: typing.List[Entity]):
        pass

    @abstractmethod
    def emit_relationships(self, relationships: typing.List[Relationship]):
        pass

    @abstractmethod
    def emit_errors(self, errors: typing.List[Error]):
        pass


class DefaultEmitter(Emitter):
    def __init__(self, context: IngestionSessionContext):
        self.ctx = context
        self.fd_map: typing.Dict[str, IO] = {}
        self.entities_lz = LandingZoneFactory.build_from_uri(
            context.entities_output_location
        )
        self.assets_lz = LandingZoneFactory.build_from_uri(
            context.assets_output_location
        )
        self.relationships_lz = LandingZoneFactory.build_from_uri(
            context.relationships_output_location
        )
        self.errors_lz = LandingZoneFactory.build_from_uri(
            context.errors_output_location
        )

    def __get_entities_fd(self, entity: Entity) -> IO:
        if "entities" not in self.fd_map:
            self.fd_map["entities"] = tempfile.NamedTemporaryFile(
                mode="w+b", delete=False, prefix="entities"
            )
        return self.fd_map["entities"]

    def __get_errors_fd(self, error: ErrorEntry) -> IO:
        if "errors" not in self.fd_map:
            self.fd_map["errors"] = tempfile.NamedTemporaryFile(
                mode="w+b", delete=False, prefix="errors"
            )
        return self.fd_map["errors"]

    def __get_relationships_fd(self, relationship: Relationship) -> IO:
        if "relationships" not in self.fd_map:
            self.fd_map["relationships"] = tempfile.NamedTemporaryFile(
                mode="w+b", delete=False, prefix="relationships"
            )
        return self.fd_map["relationships"]

    def __get_assets_fd(self, asset: Asset) -> IO:
        if asset.resource_code not in self.fd_map:
            self.fd_map[asset.resource_code] = tempfile.NamedTemporaryFile(
                mode="w+", delete=False, prefix=asset.resource_code
            )
        return self.fd_map[asset.resource_code]

    def __write_entity(self, entity: Entity):
        f = self.__get_entities_fd(entity)
        p = entity.to_proto()

        f.write(_VarintBytes(p.ByteSize()))
        f.write(p.SerializeToString())

    def __write_error(self, error: ErrorEntry):
        f = self.__get_errors_fd(error)
        p = error.to_proto()

        f.write(_VarintBytes(p.ByteSize()))
        f.write(p.SerializeToString())

    def __write_relationship(self, relationship: Relationship):
        f = self.__get_relationships_fd(relationship)
        p = relationship.to_proto()

        f.write(_VarintBytes(p.ByteSize()))
        f.write(p.SerializeToString())

    def __write_asset(self, asset: Asset):
        f = self.__get_assets_fd(asset)
        f.write(asset.to_csv_entry(self.ctx) + "\n")

    def emit_entities(self, entities: typing.List[Entity]):
        if entities is None or len(entities) <= 0:
            raise ValueError("Entities cannot be null or empty.")

        for e in entities:
            self.__write_entity(e)

    def emit_relationships(self, relationships: typing.List[Relationship]):
        if relationships is None or len(relationships) <= 0:
            raise ValueError("Relationships cannot be null or empty.")

        for r in relationships:
            self.__write_relationship(r)

    def emit_errors(self, errors: typing.List[Error]):
        if errors is None or len(errors) <= 0:
            raise ValueError("Errors cannot be null or empty.")

        for e in errors:
            self.__write_error(e)

    def emit_assets(self, assets: typing.List[Asset]):
        if assets is None or len(assets) <= 0:
            raise ValueError("Assets cannot be null or empty.")

        for a in assets:
            self.__write_asset(a)

    def close(self):
        for k, v in self.fd_map.items():
            v.close()

            if k == "entities":
                self.entities_lz.upload(
                    v.name,
                    self.ctx.entities_output_location,
                    ".proto.gz",
                    prefix="entities",
                )
            elif k == "relationships":
                self.relationships_lz.upload(
                    v.name,
                    self.ctx.relationships_output_location,
                    ".proto.gz",
                    prefix="relationships",
                )
            elif k == "errors":
                self.errors_lz.upload(
                    v.name,
                    self.ctx.errors_output_location,
                    ".proto.gz",
                    prefix="errors",
                )
            else:
                self.assets_lz.upload(
                    v.name,
                    self.ctx.assets_output_location + k + "/",
                    ".csv.gz",
                )



import datetime
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

from sqlalchemy.sql import FromClause
from sqlalchemy.sql.elements import BooleanClauseList

from bakplane.bakplane_pb2 import (
    Universe,
    ResolveResourceIntentResponse,
    ResolveResourcePathResponse,
)

from sqlalchemy import Table, Column


class WriteMode(Enum):
    ERROR_IF_EXISTS = 1
    APPEND = 2
    OVERWRITE = 3
    IGNORE = 4


@dataclass
class ExecutionStatistics:
    elapsed_ms: float


@dataclass
class ConnectionStringContext:
    context: typing.Dict[str, str]


@dataclass
class WriteRequest:
    path: str
    warehouse: str
    df: typing.Any
    mode: WriteMode


@dataclass
class ReadAllRequest:
    path: str
    warehouse: str
    columns: typing.List[str]


@dataclass
class ReadRequestBuildingBlocks:
    asset_table: Table
    mapping_table: Table
    selectable: FromClause
    constraints: typing.List[typing.Any]
    effective_start_dt: datetime.datetime = None
    effective_end_dt: datetime.datetime = None
    knowledge_start_dt: datetime.datetime = None
    knowledge_end_dt: datetime.datetime = None
    query: typing.Any = None
    columns: typing.List[Column] = None
    modifier_context: typing.Dict[str, typing.Any] = None


class DataStoreHelper(ABC):
    @abstractmethod
    def create_temporary_table(
        self,
        prefix: str,
        columns: typing.List[Column],
        rows: typing.List[typing.Any],
    ) -> Table:
        pass


class Modifier:
    @abstractmethod
    def execute(
        self, r: ReadRequestBuildingBlocks, h: DataStoreHelper
    ) -> ReadRequestBuildingBlocks:
        pass


@dataclass
class ReadRequest:
    path: str
    warehouse: str
    universe: Universe
    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime
    pointers: typing.Dict[str, str]
    columns: typing.List[str] = None
    knowledge_start_dt: datetime.datetime = None
    knowledge_end_dt: datetime.datetime = None
    query_builder: typing.Callable[
        [ReadRequestBuildingBlocks], typing.Any
    ] = None
    additional_constraints: typing.Callable[
        [ReadRequestBuildingBlocks], BooleanClauseList
    ] = None
    effective_dating_hint: typing.Union[
        str, typing.Tuple[datetime.datetime, datetime.datetime]
    ] = None
    modifiers: typing.List[Modifier] = None


@dataclass
class ReadUniverseRequest:
    universe: Universe
    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime
    knowledge_start_dt: datetime.datetime = None
    knowledge_end_dt: datetime.datetime = None


@dataclass
class ReadResponse:
    df: typing.Any
    execution_statistics: ExecutionStatistics
    context: typing.Dict[str, str]


@dataclass
class WriteResponse:
    execution_statistics: ExecutionStatistics


@dataclass
class ReadUniverseResponse:
    df: typing.Any
    execution_statistics: ExecutionStatistics


@dataclass
class PluginEntry:
    name: str
    code: str
    description: str
    author: str
    write_fn: typing.Callable[
        [WriteRequest, ResolveResourcePathResponse], WriteResponse
    ]
    read_fn: typing.Callable[
        [typing.Any, ReadRequest, ResolveResourceIntentResponse], ReadResponse
    ]
    read_all_fn: typing.Callable[
        [typing.Any, ReadAllRequest, ResolveResourcePathResponse], ReadResponse
    ]
    connection_string_url_fn: typing.Callable[[ConnectionStringContext], str]
    driver_name: str


class BaseExtension(ABC):
    @abstractmethod
    def read(self, r: ReadRequest) -> ReadResponse:
        pass

    @abstractmethod
    def read_all(self, r: ReadAllRequest) -> ReadResponse:
        pass

    @abstractmethod
    def write(self, r: WriteRequest) -> WriteResponse:
        pass

    @abstractmethod
    def read_universe(self, r: ReadUniverseRequest) -> ReadUniverseResponse:
        pass



#!/usr/bin/env python

import json
import os
import pathlib
import typing
from abc import ABC, abstractmethod
from functools import lru_cache

import boto3
import click
from sqlalchemy import PrimaryKeyConstraint
from sqlalchemy import Table, Column, String, MetaData, TIMESTAMP, BigInteger, Index
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql.base import BYTEA
from sqlalchemy.schema import CreateTable
from sqlalchemy.sql.ddl import CreateIndex
from sqlalchemy.sql.sqltypes import BOOLEAN, DATE, DATETIME, DECIMAL, FLOAT, Integer, JSON, TIME
from sqlalchemy_redshift.dialect import RedshiftDialect

UP_INITIAL_MIGRATION_NAME = '1_initial.up.sql'
TABLES_MIGRATION_NAME = '2_tables.up.sql'


class MigrationProvider(ABC):
    def get_dating_columns(self):
        return ['effective_start_dt', 'effective_end_dt', 'knowledge_start_dt', 'knowledge_end_dt']

    def get_hashing_columns(self):
        return ['knowledge_hash', 'payload_hash']

    def supports_partitioning(self):
        return False

    def supports_indices(self):
        return False

    def get_partition_ranges(self):
        return [1900] + list(range(1995, 2030)) + [3000]

    def get_metadata(self) -> MetaData:
        return MetaData()

    @abstractmethod
    def get_sqlalchemy_dialect(self):
        pass

    @abstractmethod
    def get_name(self):
        pass

    def add_partitioning_constraint(self, tbl: Table):
        pass

    def get_auxiliary_columns(self):
        return [
            Column('ingestion_session_id', BigInteger, nullable=False),
            Column('resource_code', String(128), nullable=False)
        ]

    def create_partition_table(self, f, schema: str, parent_name: str, _from: str, to: str):
        pass


class PostgresProvider(MigrationProvider):
    def supports_indices(self):
        return True

    def supports_partitioning(self):
        return True

    def get_sqlalchemy_dialect(self):
        return postgresql.dialect()

    def get_name(self):
        return 'postgres'

    def create_partition_table(self, f, schema: str, parent_name: str, _from: str, to: str):
        stmt = f"CREATE TABLE IF NOT EXISTS {schema}.{parent_name}_p{_from}_{to} PARTITION OF {schema}.{parent_name} FOR VALUES FROM ('{_from}-01-01') TO ('{to}-01-01');\n"
        f.write(stmt)

    def add_partitioning_constraint(self, tbl: Table, column: str = 'effective_start_dt'):
        tbl.dialect_options['postgresql']['partition_by'] = f'RANGE({column})'


class RedshiftProvider(MigrationProvider):
    def supports_indices(self):
        return False

    def supports_partitioning(self):
        return False

    def get_sqlalchemy_dialect(self):
        return RedshiftDialect()

    def get_name(self):
        return 'redshift'


class MigrationProviderFactory(object):
    @staticmethod
    @lru_cache()
    def get(warehouse):
        if warehouse == 'postgres' or warehouse == 'postgresql':
            return PostgresProvider()
        elif warehouse == 'redshift':
            return RedshiftProvider()
        raise ValueError('Invalid warehouse provided.')


def create_schema_if_needed(f, provider: MigrationProvider, doc: typing.Dict[str, typing.Any]):
    required_schemas = [doc['provider']]

    for r in doc['resources']:
        required_schemas.append(r['dataset'])

    f.write(f'CREATE SCHEMA IF NOT EXISTS temp_data;\n')

    for rs in set(required_schemas):
        f.write(f'CREATE SCHEMA IF NOT EXISTS {rs};\n')


def create_table_if_needed(f, provider: MigrationProvider, r: typing.Dict[str, typing.Any], schema: str):
    columns = []
    indices = [
        ['knowledge_hash', 'payload_hash', 'resource_code'],
    ]

    ix_count = 0

    found_context = False
    for ctx in r['context']:
        if ctx['warehouse_kind'] == provider.get_name():
            found_context = True
            for p in ctx['pointers']:
                columns.append(Column(
                    p['column_name'],
                    BigInteger,
                    nullable=False
                ))

                indices.append([p['column_name'], 'effective_start_dt', 'effective_end_dt'])

    if not found_context:
        if 'pointers' in r and len(r['pointers']) > 0:
            for p in r['pointers']:
                columns.append(Column(
                    f'{p["entity_type"]}_identity',
                    BigInteger,
                    nullable=False
                ))

    # add atttributes
    for c in r['schema']['attributes']:
        columns.append(Column(
            c['name'],
            map_attribute_type(c['type'], c),
            nullable='mode' in c and c['mode'] == 'nullable',
        ))

    # add dating columns
    for dc in provider.get_dating_columns():
        columns.append(Column(
            dc,
            TIMESTAMP,
            nullable=False
        ))

    # add hashing
    for hc in provider.get_hashing_columns():
        columns.append(Column(
            hc,
            BigInteger,
            nullable=False
        ))

    columns.extend(provider.get_auxiliary_columns())

    tbl = Table(
        r['code'],
        provider.get_metadata(),
        *columns,
        PrimaryKeyConstraint('knowledge_hash', 'payload_hash', 'knowledge_end_dt', 'effective_start_dt'),
        schema=schema
    )

    if provider.supports_partitioning():
        provider.add_partitioning_constraint(tbl)

    tbl_s = str(CreateTable(tbl).compile(dialect=provider.get_sqlalchemy_dialect())).replace("CREATE TABLE",
                                                                                             "CREATE TABLE IF NOT EXISTS").rstrip()
    f.write(f'{tbl_s};\n')

    if provider.supports_indices():
        for i in indices:
            ix = Index(f'{r["code"]}_ix_{ix_count}', *[tbl.c.get(x) for x in i])
            ix_count += 1

            cx = str(CreateIndex(ix).compile(dialect=provider.get_sqlalchemy_dialect())).replace("CREATE INDEX",
                                                                                                 "CREATE INDEX IF NOT EXISTS")
            f.write(f'{cx};\n')

    if provider.supports_partitioning():
        ranges = provider.get_partition_ranges()
        for i in range(0, len(ranges) - 1):
            provider.create_partition_table(f, schema, r['code'], str(ranges[i]), str(ranges[i + 1]))

    return tbl


def map_attribute_type(t: str, col):
    if t == 'string':
        return String(col.get('length', 128))
    elif t == 'integer':
        return Integer
    elif t == 'numeric':
        return DECIMAL(scale=col.get('scale', 4), precision=col.get('precision', 15))
    elif t == 'timestamp':
        return TIMESTAMP
    elif t == 'boolean':
        return BOOLEAN
    elif t == 'floating':
        return FLOAT
    elif t == 'bytes':
        return BYTEA
    elif t == 'date':
        return DATE
    elif t == 'datetime':
        return DATETIME
    elif t == 'time':
        return TIME
    elif t == 'json':
        return JSON
    elif t == 'bigint':
        return BigInteger

    raise ValueError(f'Invalid field type `{t}`.')


def handle_schemas(provider: MigrationProvider, r: typing.Dict[str, typing.Any], doc: typing.Dict[str, typing.Any]):
    dest_file = os.path.join('migrations', r['code'], provider.get_name(), UP_INITIAL_MIGRATION_NAME)
    pathlib.Path(os.path.join('migrations', r['code'], provider.get_name())).mkdir(parents=True, exist_ok=True)

    click.echo(f'Adding initial migrations to `{dest_file}` ...')

    with open(dest_file, 'w') as f:
        create_schema_if_needed(f, provider, doc)

    click.echo(f'Initial migrations sucessfully written to `{dest_file}`.')
    return dest_file


def handle_tables(provider: MigrationProvider, r: typing.Dict[str, typing.Any], doc: typing.Dict[str, typing.Any]):
    dest_file = os.path.join('migrations', r['code'], provider.get_name(), TABLES_MIGRATION_NAME)
    pathlib.Path(os.path.join('migrations', r['code'], provider.get_name())).mkdir(parents=True, exist_ok=True)

    click.echo(f'Adding table migrations to `{dest_file}` ...')
    with open(dest_file, 'w') as f:
        create_table_if_needed(f, provider, r, doc['provider'])
    click.echo(f'Table migrations sucessfully written to `{dest_file}`.')

    return dest_file


def generate_initial_migration_for_resource(provider: MigrationProvider, r: typing.Dict[str, typing.Any],
                                            doc: typing.Dict[str, typing.Any], bucket: str, adapter: str):
    client = boto3.client('s3')

    schema_migration_file = handle_schemas(provider, r, doc)
    tbl_migration_file = handle_tables(provider, r, doc)

    schema_migration_dst_path = adapter + '/' + r[
        'code'] + '/' + provider.get_name() + '/' + os.path.basename(schema_migration_file)
    click.echo(f'Uploading schema migration to `{schema_migration_dst_path}.')
    client.upload_file(schema_migration_file, bucket, schema_migration_dst_path)

    tbl_migration_dst_path = adapter + '/' + r[
        'code'] + '/' + provider.get_name() + '/' + os.path.basename(tbl_migration_file)
    click.echo(f'Uploading table migration to `{tbl_migration_dst_path}.')

    client.upload_file(tbl_migration_file, bucket, tbl_migration_dst_path)


def generate_initial_migrations(provider: MigrationProvider, doc: typing.Dict[str, typing.Any], bucket: str,
                                adapter: str):
    for r in doc['resources']:
        generate_initial_migration_for_resource(provider, r, doc, bucket, adapter)


@click.command()
@click.option('--metadata', '-m', required=True, help="The metadata resource path.")
@click.option('--adapter', '-a', required=True, help="The adapter name")
@click.option('--warehouse', '-w', multiple=True, required=True,
              help="You can provide multiple warehouse destinations.")
@click.option('--bucket', '-b', default='bakplane-dev-migrations', required=True,
              help="The bucket containing the migrations.")
def main(metadata, adapter, warehouse, bucket):
    with open(metadata, 'r') as f:
        doc = json.load(f)
        for w in warehouse:
            generate_initial_migrations(MigrationProviderFactory.get(w), doc, bucket, adapter)


if __name__ == '__main__':
    main()



import datetime
import typing
import os

from dataclasses import dataclass

from dominus.dominus_pb2 import (
    EntityEntry,
    Descriptor as PDescriptor,
    IdentityEntry,
    TimestampRange,
    RelationshipEntry,
)

from bakplane.utils import to_proto_timestamp, to_timestamp, hash_string
from bakplane.bakplane_pb2 import ErrorEntry


@dataclass
class IngestionSessionContext:
    ingestion_session_id: int
    entities_output_location: str
    assets_output_location: str
    errors_output_location: str
    relationships_output_location: str

    @staticmethod
    def from_env():
        return IngestionSessionContext(
            ingestion_session_id=int(os.getenv("INGESTION_SESSION_ID", 0)),
            entities_output_location=os.getenv(
                "ENTITIES_OUTPUT_LOCATION", "file://./output/entities/"
            ),
            assets_output_location=os.getenv(
                "ASSETS_OUTPUT_LOCATION", "file://./output/assets/"
            ),
            errors_output_location=os.getenv(
                "ERRORS_OUTPUT_LOCATION", "file://./output/errors/"
            ),
            relationships_output_location=os.getenv(
                "RELATIONSHIPS_OUTPUT_LOCATION",
                "file://./output/relationships/",
            ),
        )


@dataclass
class Asset:
    resource_code: str
    retrieved_by: typing.Dict[str, int]
    fields: typing.Dict[str, str]

    key_hash_fields: typing.List[str]
    payload_hash_fields: typing.List[str]

    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime
    knowledge_start_dt: datetime.datetime = None
    knowledge_end_dt: datetime.datetime = None

    def calculate_hash(self, f: typing.List[str]):
        parts = set()

        for k in f:
            if k in self.fields:
                parts.add(str(self.fields[k]))
            elif k in self.retrieved_by:
                parts.add(str(self.retrieved_by[k]))
            elif k == "effective_start_dt":
                parts.add(str(to_timestamp(self.effective_start_dt)))
            elif k == "effective_end_dt":
                parts.add(str(to_timestamp(self.effective_end_dt)))
            elif k == "knowledge_start_dt":
                if self.knowledge_start_dt:
                    parts.add(str(to_timestamp(self.knowledge_start_dt)))
            elif k == "knowledge_end_dt":
                if self.knowledge_end_dt:
                    parts.add(str(to_timestamp(self.knowledge_end_dt)))

        return hash_string("".join(sorted(list(parts))))

    @property
    def key_hash(self):
        return self.calculate_hash(self.key_hash_fields)

    @property
    def payload_hash(self):
        return self.calculate_hash(self.payload_hash_fields)

    def to_csv_entry(self, ctx: IngestionSessionContext):
        parts = []

        sorted_retrieved_by = sorted(self.retrieved_by.keys())
        sorted_fields = sorted(self.fields.keys())

        for k in sorted_retrieved_by:
            parts.append(str(self.retrieved_by[k]))

        for k in sorted_fields:
            parts.append(str(self.fields[k]))

        parts.append(
            self.effective_start_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )
        parts.append(
            self.effective_end_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )

        if self.knowledge_start_dt is None:
            self.knowledge_start_dt = datetime.datetime.utcnow()

        if self.knowledge_end_dt is None:
            self.knowledge_end_dt = datetime.datetime(
                year=3000, month=12, day=31
            )

        parts.append(
            self.knowledge_start_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )
        parts.append(
            self.knowledge_end_dt.strftime("%Y-%m-%d") + " 00:00:00.00000"
        )

        parts.append(self.resource_code)
        parts.append(str(ctx.ingestion_session_id))

        parts.append(str(self.key_hash))
        parts.append(str(self.payload_hash))

        return "|".join(parts)


@dataclass
class Identity:
    entity_type: str
    perspective: str
    attributes: typing.Dict[str, str]

    @property
    def uid(self):
        c = []
        for i in sorted(self.attributes.keys()):
            c.append(i + self.attributes[i])

        return hash_string(
            "".join([self.perspective, self.entity_type, "".join(c)])
        )

    def to_proto(self) -> IdentityEntry:
        return IdentityEntry(
            uid=self.uid,
            entity_type_code=self.entity_type,
            perspective_code=self.perspective,
        )


@dataclass
class Descriptor:
    attribute: str
    value: str
    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime
    knowledge_start_dt: datetime.datetime = None
    knowledge_end_dt: datetime.datetime = None

    def to_proto(self) -> PDescriptor:
        return PDescriptor(
            attribute_code=self.attribute,
            value=self.value,
            effective_dating=[
                TimestampRange(
                    start_dt=to_proto_timestamp(self.effective_start_dt),
                    end_dt=to_proto_timestamp(self.effective_end_dt),
                )
            ],
        )


@dataclass
class Entity:
    identity: Identity
    descriptors: typing.List[Descriptor]

    def to_proto(self) -> EntityEntry:
        return EntityEntry(
            identity=self.identity.to_proto(),
            descriptors=[x.to_proto() for x in self.descriptors],
        )


@dataclass
class Relationship:
    source_identity_uid: int
    target_identity_uid: int
    relationship_type: str
    effective_start_dt: datetime.datetime
    effective_end_dt: datetime.datetime

    def to_proto(self) -> RelationshipEntry:
        return RelationshipEntry(
            source_identity_uid=self.source_identity_uid,
            target_identity_uid=self.target_identity_uid,
            relationship_type_code=self.relationship_type,
            effective_dating=[
                TimestampRange(
                    start_dt=to_proto_timestamp(self.effective_start_dt),
                    end_dt=to_proto_timestamp(self.effective_end_dt),
                )
            ],
        )


@dataclass
class Error:
    message: str
    stacktrace: str = None
    context: typing.Dict[str, str] = None

    def to_proto(self) -> ErrorEntry:
        return ErrorEntry(
            message=self.message,
            stacktrace=self.stacktrace,
            context=self.context,
        )



import gzip
import os
import shutil
import tempfile
from abc import ABC, abstractmethod
from urllib.parse import urlparse

import boto3


class LandingZone(ABC):
    @abstractmethod
    def upload(self, src_path: str, dst_path: str, suffix: str = "", prefix: str = None):
        pass


class FilesystemLandingZone(LandingZone):
    def upload(
        self,
        src_path: str,
        dst_path: str,
        suffix: str = "",
        prefix: str = None,
    ):
        u = urlparse(dst_path)
        with open(src_path, "rb") as f_in:
            with tempfile.NamedTemporaryFile(
                "wb",
                prefix=prefix,
                delete=False,
                dir="." + u.path,
                suffix=suffix,
            ) as f:
                with gzip.open(f, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)


class AmazonS3LandingZone(LandingZone):
    def __init__(self):
        self.client = boto3.client("s3")

    def upload(self, src_path: str, dst_path: str, suffix: str = "", prefix: str = None):
        u = urlparse(dst_path)

        if u.scheme != "s3":
            raise RuntimeError("Invalid destination path.")

        with open(src_path, "rb") as f_in:
            f = tempfile.NamedTemporaryFile(delete=False)
            final_path = f.name

            with gzip.open(f, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

            f.close()

            self.client.upload_file(
                final_path,
                u.netloc,
                u.path[1:] + os.path.basename(src_path) + suffix,
            )


class LandingZoneFactory(object):
    @staticmethod
    def build_from_uri(uri):
        u = urlparse(uri)

        if u.scheme == "s3":
            return AmazonS3LandingZone()
        elif u.scheme == "file":
            return FilesystemLandingZone()
        raise RuntimeError(
            f"Could not find landing zone matching scheme `{u.scheme}`."
        )
